/* 
 * Copyright 2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cstdio>
#include <sstream>
#include <iomanip>
#include "PBStreamer.h"
#include "spoof.h"

PBStreamer::Result PBStreamer::writeMessage(
    ::google::protobuf::MessageLite *msg)
{
    if (!wrotemagic && _magic) {
#ifdef _WIN32
	// send() works only on sockets (on all platforms)
	unsigned n = send(_fd, _magic, _magiclen, 0);
#else
	// on Unix, write() works on all types of file descriptors
	unsigned n = write(_fd, _magic, _magiclen);
#endif
	if (n != _magiclen)
	    return IO_ERROR;
	wrotemagic = true;
	if (readmagic) {
	    delete _magic; // no longer needed;
	    _magic = 0;
	}
    }

    if (msg_size == 0) {
	msg_size = msg->ByteSize();
	if (capacity < sizeof(msg_size) + msg_size) {
	    if (buf) delete buf;
	    buf = new char[sizeof(msg_size) + msg_size];
	    if (!buf) {
		capacity = 0;
		return cleanup(OUT_OF_MEMORY);
	    }
	    capacity = sizeof(msg_size) + msg_size;
	}
	*(uint32_t*)buf = htonl(msg_size);
	msg->SerializeWithCachedSizesToArray((uint8_t*)buf + sizeof(msg_size));
	progress = 0;
    }
    Result result = writeBytes(sizeof(msg_size) + msg_size);
    return cleanup(result, false);
}

PBStreamer::Result PBStreamer::readMessage(
    ::google::protobuf::MessageLite *msg)
{
    Result result;

    if (!readmagic && _magic) {
	if (capacity < _magiclen) {
	    buf = new char[_magiclen];
	    if (!buf) {
		capacity = 0;
		return cleanup(OUT_OF_MEMORY);
	    }
	    capacity = sizeof(msg_size);
	}
	result = readBytes(_magiclen);
	if (result != OK)
	    return cleanup(result, false);
	if (memcmp(_magic, buf, _magiclen) != 0) {
	    std::ostringstream ess;
	    ess << std::right << std::hex << std::setfill('0') << "bad magic:";
	    for (unsigned i = 0; i < _magiclen; i++)
		ess << ' ' << std::setw(2) << (unsigned)buf[i];
	    if (isprint(buf[0])) {
		ess << " (\"";
		for (unsigned i = 0; i < _magiclen; i++)
		    ess << (isprint(buf[i]) ? buf[i] : '.');
		ess << "\")";
	    }
	    errstr = ess.str();
	    return cleanup(BAD_MAGIC, false);
	}
	readmagic = true;
	if (wrotemagic) {
	    delete _magic; // no longer needed;
	    _magic = 0;
	}
	progress = 0;
    }

    if (capacity < sizeof(msg_size)) {
	buf = new char[sizeof(msg_size)];
	if (!buf) {
	    capacity = 0;
	    return cleanup(OUT_OF_MEMORY);
	}
	capacity = sizeof(msg_size);
    }
    if (msg_size == 0) {
	result = readBytes(sizeof(msg_size));
	if (result != OK)
	    return cleanup(result, false);
	msg_size = ntohl(*(uint32_t*)buf);
	// Sanity check: a maximum length of 0x00FFFFFF will block any non-PBS
	// stream that starts with an ASCII character or other nonzero byte,
	// but still allow PBS messages up to 16 MiB.
	if (msg_size > 0x00FFFFFF)
	    return cleanup(PARSE_ERROR);
	progress = 0;
    }
    if (capacity < msg_size) {
	if (buf) delete buf;
	buf = new char[msg_size];
	if (!buf) {
	    capacity = 0;
	    return cleanup(OUT_OF_MEMORY);
	}
	capacity = msg_size;
    }
    result = readBytes(msg_size);
    if (result != OK)
	return cleanup(result, false);
    msg->Clear(); // XXX is this necessary?
    if (!msg->ParseFromArray(buf, msg_size))
	return cleanup(PARSE_ERROR);
    msg_size = 0;
    progress = 0;
    return OK;
}

// fills in errstr if result is not OK
PBStreamer::Result PBStreamer::writeBytes(uint32_t goal)
{
    ssize_t n;

    if (progress == goal) return OK;
#ifdef _WIN32
    // send() works only on sockets (on all platforms)
    n = send(_fd, buf + progress, goal - progress, 0);
#else
    // on Unix, write() works on all types of file descriptors
    n = write(_fd, buf + progress, goal - progress);
#endif
    if (n >= 0) {
	progress += n;
	if (progress == goal) return OK; // finished
	return set_errbuf(INCOMPLETE);
#ifndef _WIN32 // nonblocking not supported on Windows
    } else if (errno == EAGAIN || errno == EWOULDBLOCK) {
	return set_errbuf(INCOMPLETE);
#endif
    } else {
	return set_errbuf(IO_ERROR);
    }
}

// fills in errstr if result is not OK
PBStreamer::Result PBStreamer::readBytes(uint32_t goal)
{
    int n;

    if (progress == goal) return OK;
#ifdef _WIN32
    // recv() works only on sockets (on all platforms)
    n = recv(_fd, buf + progress, goal - progress, 0);
#else
    // on Unix, read() works on all types of file descriptors
    n = read(_fd, buf + progress, goal - progress);
#endif

#if DEBUG-0
    if (n < 0)
	fprintf(stderr, "read: %s\n", sock_last_errmsg());
    else
	fprintf(stderr, "read %d of %d bytes\n", n, goal - progress);
#endif

    if (n == 0) {
	return set_errbuf(CLOSED); // peer closed
    } else if (n > 0) {
	progress += n;
	if (progress == goal) return OK; // finished
	return set_errbuf(INCOMPLETE);
#ifndef _WIN32 // nonblocking not supported on Windows
    } else if (errno == EAGAIN || errno == EWOULDBLOCK) {
	return set_errbuf(INCOMPLETE);
#endif
    } else {
	return set_errbuf(IO_ERROR);
    }
}

const char *PBStreamer::errmsg(PBStreamer::Result result)
{
    static char buf[1024];
    switch (result) {
    case OK:            return "no error";
    case INCOMPLETE:    return "operation is incomplete";
    case CLOSED:        return "remote host closed connection";
    case OUT_OF_MEMORY: return "out of memory";
    case IO_ERROR:      sprintf(buf, "I/O error: %s", sock_last_errmsg());
			return buf;
    case BAD_MAGIC:     return "bad magic";
    case PARSE_ERROR:   return "protocol parse error";
    default:            sprintf(buf, "unknown error %d", result);
			return buf;
    }
}
