/* 
 * Copyright 2004-2009 Rob Beverly
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

/****************************************************************************
   Program:     $Id: util.cc,v 1.28 2016/06/02 17:27:36 kkeys Exp $
   Author:      Rob Beverly <rbeverly at csail.mit.edu>
   Date:        $Date: 2016/06/02 17:27:36 $
   Description: Spoofer utility routines
****************************************************************************/
#include "spoof.h"
#include <stdarg.h>
#include <ctype.h>

int verbosity = OFF;


// snprintf() in (some versions of?) Windows does not match the C standard:
// - on overflow, doesn't write the terminating NUL
// - on overflow, returns -1 instead of the would-be string length
// This version adds the terminating NUL, but does not fix the return value
// (that would be much harder to fix, and we don't need that feature anyway).
// This version works whether the wrapped vsnprintf() follows the standard or
// Windows.
int spoofer_snprintf(char *buf, size_t size, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    int n = vsnprintf(buf, size, fmt, ap);
    va_end(ap);
    buf[size-1] = '\0';
    return n;
}

int is_interactive(void)
{
#ifdef _WIN32
    HANDLE h = GetStdHandle(STD_INPUT_HANDLE); 
    if (h == INVALID_HANDLE_VALUE) return 0;
    DWORD dummyMode;
    return GetConsoleMode(h, &dummyMode);
#else
    return isatty(STDIN_FILENO) && isatty(STDOUT_FILENO);
#endif
}

NORETURN exitpause(int status) {
#ifdef _WIN32
    DWORD proclist[1];
    if (is_interactive() && GetConsoleProcessList(proclist, 1) <= 1) {
	// This console has no other processes; it will close when we exit.
	fprintf(stdout, "\nPress Enter to Exit.\n"); \
	fflush(stdout); \
	while (getchar() != '\n');
    }
#endif
    exit(status);
}

/**
 * Fatal error condition function 
 */
NORETURN fatal(const char *fmt, ...) {
    va_list ap;
    fflush(stdout);
    va_start(ap, fmt);
    fprintf(stderr, "*** ");
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exitpause(-1);
}

/**
 * Fatal error condition function; prints error
 */
NORETURN pfatal(const char *fmt, ...) {
    va_list ap;
    fflush(stdout);
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, " ");
    perror("Error");
    va_end(ap);
    exitpause(-1);
}

/**
 * binDump -  Debug routine for examining raw packets/buffer/char strings 
 *
 * @param[in]   msg     pointer to message buffer
 * @param[in]   len     length of buffer to print
 * @param[in]   pretty  if true, pretty print
 */
void binDump(const unsigned char *msg, int len, int pretty) {
    int i;
    for (i=1;i<=len;i++) {
	if (pretty) {
	    printf("%02x ", *msg++);
	    if (i > len) break;
	    if (i%16 == 0) printf("\n");
	    else if (i%4 == 0) printf(" ");
	} else {
	    printf("%02x", *msg++);
	}
    }
    if (pretty) printf("\n");
}

/**
 * byte32Swap
 */
unsigned long byte32Swap(unsigned long in) {
    return ( (in & 0x000000ff)<<24 | (in & 0x0000ff00)<<8 |
             (in & 0x00ff0000)>>8 | (in & 0xff000000)>>24 );
}

/**
 * byte16Swap
 */
unsigned short byte16Swap(unsigned short in) {
    return ( (in & 0x00ff)<<8 | (in & 0xff00)>>8 );
}

/**
 * win_init -  Windows socket API startup 
 */
void win_init(void) {
#ifdef _WIN32
    WSADATA wsd;
    if (WSAStartup(MAKEWORD(2, 2), &wsd) != 0) {
        fatal("WSAStartup() failed");
    }
#endif
}

/**
 * msleep -  Portable millisecond sleep 
 */
void msleep(int msec) {
#ifdef _WIN32
    Sleep(msec);
#else
    usleep(msec * 1000);
#endif
}

/**
 * spoofSleep -  Portable 0.1 - 0.5 s random sleep 
 */
void spoofSleep(void) {
    msleep((unsigned int)(rand() / ((double)RAND_MAX + 1) * 400) + 100);
}

/**
 * timestring - form time/date string
 */
char *timestring(void) {
   static char str[BUFSIZE];
   time_t now;
   struct tm *t;

   now = time(NULL);
   t = localtime(&now);
   snprintf(str, sizeof(str), "%d/%02d/%02d %02d:%02d:%02d", t->tm_year + 1900,
       t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec);
   return (str);
}

/**
 * streamSetup - set stdout to be unbuffered, replicate stdout
 *               to the stderr file descriptor.  useful with
 *               the authorization services API (for osx gui)
 *               which doesn't pipe back stderr and prevents
 *               weird interleaving of stderr, stdout messages.
 */
void streamSetup(void) {
    if (setvbuf(stdout, NULL, _IONBF, 0) == EOF) {
        perror("setvbuf");
    }
    if (dup2(1,2) < 0) {
        perror("dup2");
    }
}

/* Unified IPv4/IPv6 address field helper routines.
 *   Ex:
 *     char bob[IPV6ADDRLEN];
 *     uint32_t alice = 91232354;
 *     h4to6n(alice, bob);
 *     printf("Alice: %u %u\n", alice, n6toh4(bob));
 */
/* network byte-order IPv6 to network byte-order IPv4 */
uint32_t n6ton4(const unsigned char *addr) {
    const unsigned char *begin = addr + (IPV6ADDRLEN - IPV4ADDRLEN);
    uint32_t res;
    memcpy(&res, begin, IPV4ADDRLEN);
    return (res);
}

/* network byte-order IPv6 to host byte-order IPv4 */
uint32_t n6toh4(const unsigned char *addr) {
    return (ntohl(n6ton4(addr)));
}

/* network byte-order IPv4 to network byte-order IPv6 */
int n4ton6(const uint32_t ip, unsigned char *addr) {
    unsigned char *begin = addr + (IPV6ADDRLEN - IPV4ADDRLEN);
    memcpy(begin, &ip, IPV4ADDRLEN);
    return TRUE;
}

/* host byte-order IPv4 to network byte-order IPv6 */
int h4ton6(const uint32_t ip, unsigned char *addr) {
    unsigned char *begin = addr + (IPV6ADDRLEN - IPV4ADDRLEN);
    uint32_t res = htonl(ip);
    memcpy(begin, &res, IPV4ADDRLEN);
    return TRUE;
}

int n6tohexstring(const unsigned char *addr, unsigned char *res) {
    int i=0;
    snprintf((char *)res, 3, "0x");
    for (i=1;i<=IPV6ADDRLEN;i++) {
        snprintf((char *)res + (i*2), 3, "%02x", *addr);
        addr++;
    }
    return TRUE;
}

int isIPv4(const unsigned char *addr) {
    int i=0;
    const unsigned char *d = addr;
    for (i=0;i<(IPV6ADDRLEN - IPV4ADDRLEN);i++) {
        if (*d++ != 0x00)
            return FALSE;
    }
    return TRUE;
}

int isIPv6(const unsigned char *addr) {
    return (isIPv4(addr) ? FALSE : TRUE);
}

int isIPv6Equal(const unsigned char *a1, const unsigned char *a2) {
    int ret = memcmp(a1, a2, IPV6ADDRLEN);
    return (ret == 0 ? TRUE : FALSE);
}

char *printIPv6(const unsigned char *a) {
    static char v6string[SMALLBUF];
    inet_ntop(AF_INET6, a, v6string, SMALLBUF);
    return v6string;
}

// portable version of BSD's strnstr()
const char *strstrPortable(const char *a, const char *b, size_t n) {
    size_t alen, blen, i;
    // don't depend on there being a \0 within the first n chars
    for (alen = 0; alen < n; alen++)
	if (a[alen] == '\0') break;
    blen = strlen(b);
    if (blen > alen) return NULL;
    if (n > alen - blen)
	n = alen - blen;
    for (i = 0; i <= n; i++) {
	if (memcmp(a+i, b, blen) == 0)
	    return a+i;
    }
    return NULL;
}
