/* 
 * Copyright 2004-2009 Rob Beverly
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

/****************************************************************************
   Author:      Rob Beverly <rbeverly at csail.mit.edu>
                Ken Keys, CAIDA
   Date:        $Date: 2016/07/07 21:17:02 $
   Description: Spoofer - generate spoofed packets
****************************************************************************/
#include <ctype.h>
#include <string>
#include <sstream>
#include <vector>
#include "spoof.h"
#include "spoof_pcap.h"
#include "prober.h"
#include "PBStreamer.h"

static const char cvsid[] USED = "$Id: spoofer-prober.cc,v 1.110 2016/07/07 21:17:02 kkeys Exp $";

static int sharePublic = -1; // initially unknown
static int shareRemedy = -1; // initially unknown
static bool useDevServer = false;
static bool pretendMode = false;
static bool standaloneMode = false;
static const char *standaloneTraceSrc = 0;
static std::string sessionkey("");

static const char *server4 = PROD_SERVER;
static const char *server6 = PROD_SERVER6;
static const char *report_host = PROD_REPORT_HOST;

enum SessionStatus { SS_FAIL, SS_UNTESTED, SS_SUCCESS };

#if 0
#define to_str(val) \
    (static_cast<std::ostringstream&>(std::ostringstream() << val).str())
#endif

struct Stage; // forward declaration

struct Session {
    const IPv ipv;
    const int family;
    uint32_t sessionid;
    std::string int_clientip; // what I think my address is
    std::string ext_clientip; // what server thinks my address is
    int ctrlsock;
    int udpsock;
    int spoofsock;
    PBStreamer *pbs;
    SpooferClientMsg req; // The request being prepared to send to server
    SpooferServerMsg resp; // The most recent response from server
    bool rcvdHello;
    int cost, progress;
    std::vector<Stage> stages;
    int stageid;
    int schedulesDone;

    Session(IPv _ipv) :
	ipv(_ipv), family(ipvtofamily(_ipv)),
	sessionid(0),
	ctrlsock(-1), udpsock(-1), spoofsock(-1),
	pbs(0), rcvdHello(false),
	cost(0), progress(0), stages(), schedulesDone(0)
	{}
    ~Session() {
	if (pbs) delete pbs;
	if (udpsock >= 0) close(udpsock);
	if (spoofsock >= 0) close(spoofsock);
	if (ctrlsock >= 0) close(ctrlsock);
    }

    void printHeader(const char *label, int count, const char *itemdesc) {
	// NB: GUI depends on this output format
	printf("\n>> Running IPv%d test %d:  %s (%d %s%s)\n",
	    ipv, schedulesDone+1, label, count, itemdesc, count > 1 ? "s" : "");
    }
    unsigned int addrlen() { return ::addrlen(family); }
    bool init();
    bool initSockets(bool spoof, const struct sockaddr *dst);
    bool closeSocket(int &sock);
    void resetSockets();
    bool serverConnect();
    void generateFakeHello();
    bool handleTextMessage(const SpooferTextMessage &txtmsg);
    bool validateServerMsg();
    void runSpoofTests(const SpooferSpoofSchedule &sched);
    void runTracefilterTests(const SpooferTracefilterSchedule &sched);
    void sendTracefilter(const char *srcaddr, const char *dstaddr, int pathdist);
    void runTracerouteTests(const SpooferTracerouteSchedule &sched);
    bool run();
    bool pretend();
};

static inline int saneTracefilterDist(int dist)
{
    return (dist >= 1 && dist <= TTL_OUT) ? dist : 5;
}

struct Stage : public Tracker {
    Session *session;
    int cost, progress;

    Stage(Session *_session, const SpooferSpoofSchedule &sched) :
	session(_session), progress(0)
    {
	cost = sched.item_size() * DUPLICATES;
	session->cost += cost;
    }

    Stage(Session *_session, const SpooferTracefilterSchedule &sched) :
	session(_session), cost(0), progress(0)
    {
	for (int i = 0; i < sched.item_size(); i++)
	    cost += saneTracefilterDist(sched.item(i).dist()) * DUPLICATES;
	session->cost += cost;
    }

    Stage(Session *_session, const SpooferTracerouteSchedule &sched) :
	session(_session), progress(0)
    {
	cost = sched.item_size() * TRACEROUTE_ITEM_COST;
	session->cost += cost;
    }

    // NB: GUI depends on this output format
    void showProgress(int inc = 1) {
	if (progress + inc > cost)
	    inc = cost - progress;
	progress += inc;
	session->progress += inc;
	for (int i = 0; i < inc; i++)
	    printf("%d\b.", session->ipv);
    }
};

bool Session::handleTextMessage(const SpooferTextMessage &txtmsg)
{
    const char *type;
    bool result = true;
    switch (txtmsg.level()) {
	case SpooferTextMessage::ERROR:
	    type = "error";
	    result = false;
	    break;
	case SpooferTextMessage::WARNING:
	    type = "warning";
	    break;
	case SpooferTextMessage::NOTICE:
	default:
	    type = "notice";
	    break;
    }
    printf("*** IPv%d server %s: %s\n", ipv, type, txtmsg.body().c_str());
    return result;
}

bool Session::pretend() {
    if (!pretendMode) return false;
    for (int i = 0; i < stages[stageid].cost; i++) {
	msleep(200);
	stages[stageid].showProgress();
    }
    printf("\n");
    return true;
}

// UDP errors may be asynchronously generated due to an ICMP response and not
// reported until the next use of the socket.  We must open a new socket for
// each destination to be sure it doesn't receive errors from a previous probe.
bool Session::initSockets(bool spoof, const struct sockaddr *dst)
{
    {
	// We need udpsock...
	// if not spoofing, to send a nonspoofed probe;
	// if spoofing, we may need to learn the outbound iface for sniffing;
	// if spoofing layer 2, we may need to send a nonspoofed pilot packet.
	if ((udpsock = newSocket(family, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	    return false;
	debug(DEVELOP, "udpsock=%d\n", udpsock);
	int ttl = spoof ? 1 : // pilot pkt goes just far enough to be sniffed
	    TTL_OUT; // pkt must reach server and have known initial TTL
	if (setSockTTL(family, udpsock, ttl) < 0) {
	    severe("set socket TTL: %s\n", sock_last_errmsg());
	    resetSockets();
	    return false;
	}
    }
    if (udpsock >= 0 && dst) {
	// code that depends on this connect():
	// - getsockname() in new_probe_info()
	// - send() in socketSend() for nonspoof or pilot
	if (connect(udpsock, dst, sa_len(dst)) < 0) {
	    info("connect %s %d: %s", sa_ntop(dst, NULL, 0),
		ntohs(sa_port(dst)), sock_last_errmsg());
	    resetSockets();
	    return false;
	}
	debug(DEVELOP, "connect %s %d: %s\n", sa_ntop(dst, NULL, 0),
	    ntohs(sa_port(dst)), "success");
    }
    if (!spoof)
	return true;
    if (ipv == IPv4 && (spoof_layer[ipv] == 0 || spoof_layer[ipv] == 3)) {
	// we want to at least try to spoof at layer 3 
	if ((spoofsock = newSocket(family, SOCK_RAW, IPPROTO_UDP)) < 0) {
	    severe("Spoofer needs root/admin or other special privileges to "
		"open raw sockets.  See http://%s/faq.xml for details.\n",
		report_host);
	} else {
	    debug(DEVELOP, "spoofsock=%d\n", spoofsock);
	    if (optSocketHDRINCL(spoofsock) == 0) // IPv4 only
		return true; // layer 3 spoofing is possible
	}
    }
    if (spoof_layer[ipv] == 0 || spoof_layer[ipv] == 2) {
	// we may want to spoof at layer 2 
	if (spoof_layer2possible(ipv))
	    return true; // layer 2 spoofing is possible
	notice("No accessible nonlocal IPv%d interfaces.\n", ipv);
    }
    resetSockets();
    return false;
}

void Session::resetSockets()
{
    if (udpsock >= 0) ::close(udpsock); udpsock = -1;
    if (spoofsock >= 0) ::close(spoofsock); spoofsock = -1;
}

// Get and clear any error on the socket.  For best results, allow some time
// for an ICMP error response after the last operation before calling this.
static const char *getSockErr(int sock)
{
    int err = 0;
    socklen_t errlen = sizeof(err);
    if (getsockopt(sock, SOL_SOCKET, SO_ERROR, (char*)&err, &errlen) < 0) {
	return sock_last_errmsg();
    } else if (err != 0) {
	return sock_errmsg(err);
    }
    return NULL;
}

// Close a session socket.  If there were any asynchronous errors pending on
// the socket, report them and return false.
bool Session::closeSocket(int &sock)
{
    int err = 0;
    if (sock < 0) return true;
    const char *msg = getSockErr(sock);
    if (msg) info("close: %s", msg);
    ::close(sock);
    sock = -1;
    return err == 0;
}

void Session::runSpoofTests(const SpooferSpoofSchedule &sched)
{
    struct sockaddr_storage src_ss, dst_ss;
    struct sockaddr *src = (struct sockaddr *)&src_ss;
    struct sockaddr *dst = (struct sockaddr *)&dst_ss;
    int pktlen = 0;
    int displayCounter = 1;
    unsigned int netSessId = htonl(sessionid);
    unsigned char payload[PROBESIZE+sizeof(netSessId)];
    SpooferSpoofReport *report = req.add_report()->mutable_spoof();
    report->set_status(ERROR); // until proven otherwise

    printHeader("spoof", sched.item_size(), "source/destination pair");
    if (pretend()) return;

    memset(src, 0, sizeof(src_ss));
    memset(dst, 0, sizeof(dst_ss));
    src->sa_family = family;
    dst->sa_family = family;
    sa_port(src) = htons(SRC_PORT);
    sa_port(dst) = htons(SERV_PORT);

    for (int i = 0; i < sched.item_size(); i++) {
	const SpooferSpoofSchedule::Item &sItem = sched.item(i);
	SpooferSpoofReport::Item *rItem = report->add_item();
	rItem->set_srcip(sItem.srcip());
	rItem->set_dstip(sItem.dstip());
	rItem->set_seqno(sItem.seqno());
	rItem->set_status(SpooferSpoofReport::Item::SENDFAIL); // until proven otherwise
    }

    for (int i = 0; i < sched.item_size(); i++) {
	struct probe_info *pinfo = 0;
	int n_sent = 0, n_send_errors = 0, n_other_errors = 0;
	const SpooferSpoofSchedule::Item &item = sched.item(i);

	if (item.dstip().size() != addrlen()) {
	    stages[stageid].showProgress(DUPLICATES);
	    continue; // family disagrees with dst address (shouldn't happen)
	}

	bool spoof = (memcmp(ext_clientip.data(), item.srcip().data(), addrlen()) != 0);

	memcpy(sa_ipaddr(src), item.srcip().data(), addrlen());
	memcpy(sa_ipaddr(dst), item.dstip().data(), addrlen());

	if (!initSockets(spoof, dst))
	    goto enditem;

	if (verbosity >= HIGH) {
	    printf(">> Send %d %s packets", DUPLICATES,
		spoof ? "spoofed" : "non-spoofed");
	    if (spoof)
		printf(" from %s", sa_ntop(src, NULL, 0));
	    printf(" to %s...\n", sa_ntop(dst, NULL, 0));
	} else {
	    printf(" %d%c", displayCounter++, spoof ? 's' : 'n');
	}

	pinfo = new_probe_info(IPPROTO_UDP, spoof, udpsock, src, dst);
	if (!pinfo) {
	    stages[stageid].showProgress(DUPLICATES);
	    goto enditem;
	}

	/* send probe packets */
	craftProbe(payload, spoof, &item);

	if (spoof) {
	    unsigned char pkt[BIGBUF];
	    pktlen = craftPacket(pkt, payload, PROBESIZE, src, dst, TTL_OUT, 0);
	    for (int j = 0; j < DUPLICATES; j++) {
		stages[stageid].showProgress();
		if (n_send_errors) continue; // dups are useless after send error
		n_sent++;
		if (socketSendSpoof(spoofsock, udpsock, pkt, pktlen, pinfo) < 0) {
		    n_send_errors++;
		    continue;
		}
		spoofSleep();
		if (spoof_layer[ipv] == 3) {
		    const char *msg = getSockErr(spoofsock); // valid only on L3
		    if (msg) {
			n_other_errors++;
			info("spoof: %s", msg);
		    }
		}
	    }

	    if (!closeSocket(spoofsock))
		n_other_errors++;

	} else {
	    char errbuf[4096] = "";
	    int probesize = PROBESIZE;
	    if (ipv == IPv4) { // why is this only for ipv4?
		memcpy(payload + PROBESIZE, &netSessId, sizeof(sessionid));
		probesize += sizeof(netSessId);
	    }
	    for (int j = 0; j < DUPLICATES; j++) {
		stages[stageid].showProgress();
		if (n_send_errors) continue; // dups are useless after send error
		n_sent++;
		const char *msg = NULL;
		bool slept = false;
		if (socketSend(udpsock, payload, probesize, pinfo) < 0) {
		    msg = sock_last_errmsg();
		    n_send_errors++;
		} else {
		    spoofSleep(); slept = true;
		    msg = getSockErr(udpsock);
		    if (msg) n_other_errors++;
		}
		if (msg) {
		    if (verbosity < LOW) {
			if (strcmp(msg, errbuf) != 0) { // not a repeat
			    if (errbuf[0])
				strncat(errbuf, "; ", sizeof(errbuf) - strlen(errbuf) - 1);
			    strncat(errbuf, msg, sizeof(errbuf) - strlen(errbuf) - 1);
			}
		    } else {
			info("send to %s failed: %s", sa_ntop(dst, NULL, 0), msg);
		    }
		}
		if (!slept) spoofSleep();
	    }

	    if (!closeSocket(udpsock))
		n_other_errors++;

	    if (n_sent <= n_send_errors + n_other_errors && verbosity < LOW) {
		info("send failed (%d times): %s",
		    n_send_errors + n_other_errors, errbuf);
	    }
	}

	if (n_sent > n_send_errors + n_other_errors) {
	    report->mutable_item(i)->set_status(SpooferSpoofReport::Item::OK);
	}

    enditem:
	if (pinfo) free_probe_info(pinfo, n_send_errors >= n_sent);
	resetSockets();
    }
    report->set_status(DONE);
    printf("\n");
}

// IPv4 only
void Session::sendTracefilter(const char *srcaddr, const char *dstaddr, int pathdist)
{
    unsigned char pkt[BIGBUF];
    struct sockaddr_in src, dst;
    struct probe_info *pinfo = 0;
    unsigned char *icmpkey = 0;
    int pktlen;
    unsigned short sid_hi = 0, sid_lo = 0;

    src.sin_family = AF_INET;
    dst.sin_family = AF_INET;
    dst.sin_port = htons(SERV_PORT);
    memcpy(&src.sin_addr, srcaddr, IPV4ADDRLEN);
    memcpy(&dst.sin_addr, dstaddr, IPV4ADDRLEN);
    memset(pkt, 0, BIGBUF);

    /* we encode the low 2B of session ID into the UDP source port 
     * and the high 2B into the IPID field */
    sid_lo = sessionid & 0x0000ffff;
    sid_hi = (sessionid >> 16) & 0x0000ffff;
    src.sin_port = htons(sid_lo);
    if (!initSockets(true, (struct sockaddr*)&dst))
	goto end;
    pinfo = new_probe_info(IPPROTO_UDP, true, udpsock, (struct sockaddr*)&src,
	(struct sockaddr*)&dst);
    if (!pinfo) goto end;
    if (!(icmpkey = new unsigned char[pathdist]())) {
	severe("out of memory in %s\n", __func__);
	goto end;
    }
    for (int ttl = 1; ttl <= pathdist; ttl++) {
        /* the TTL of the packet will be encoded as the size of the spoofed
         * UDP datagram.  TTL = UDPlength - 8 */
        pktlen = craftPacket(pkt, icmpkey, ttl, (struct sockaddr *)&src,
	    (struct sockaddr *)&dst, ttl, sid_hi);
        printf(" %d", ttl);
	bool skip = false;
        for (int j = 0; j < DUPLICATES; j++) {
	    stages[stageid].showProgress();
	    if (skip) continue;
            if (socketSendSpoof(spoofsock, udpsock, pkt, pktlen, pinfo) < 0)
		skip = true; // don't try to send duplicates on error
	    else
		spoofSleep();
        }
    }
end:
    if (icmpkey) delete[] icmpkey;
    if (pinfo) free_probe_info(pinfo);
    resetSockets();
}

/* IPv4 only */
void Session::runTracefilterTests(const SpooferTracefilterSchedule &sched)
{
    SpooferTracefilterReport *report = req.add_report()->mutable_tracefilter();
    report->set_status(ERROR); // until proven otherwise
    if (sched.item_size() == 0)
	return;
    if (ipv != IPv4) return;

    printHeader("tracefilter", sched.item_size(), "source/destination pair");
    if (pretend()) return;

    if (sched.item_size() > 100) {
	warn("Bogus test list.");
	return;
    }
    for (int i = 0; i < sched.item_size(); i++) {
	printf("%d:", i+1);
	const SpooferTracefilterSchedule::Item &item = sched.item(i);
	int dist = saneTracefilterDist(item.dist());
	sendTracefilter(item.srcip().data(), item.dstip().data(), dist);
	printf("\n");
    }
    report->set_status(DONE);
}

/* IPv4 only */
void Session::runTracerouteTests(const SpooferTracerouteSchedule &sched)
{
    char *tracebuf = 0;
    struct in_addr *dests = 0;
    int testsRemain = 0, testsDone = 0;
    int batchSize = MAX_TRACEROUTE_CHILD;
    SpooferTracerouteReport *report = req.add_report()->mutable_traceroute();
    report->set_status(ERROR); // until proven otherwise
    if (ipv != IPv4) goto done;

    printHeader("traceroute", sched.item_size(), "destination");
    if (pretend()) goto done;

    /* create destination list and empty report items */
    if (!(dests = new struct in_addr[sched.item_size()])) {
	severe("out of memory in %s\n", __func__);
	goto done;
    }
    for (int i = 0; i < sched.item_size(); i++) {
	memcpy(&dests[i], sched.item(i).dstip().data(), addrlen());
	const SpooferTracerouteSchedule::Item &sItem = sched.item(i);
	SpooferTracerouteReport::Item *rItem = report->add_item();
	rItem->set_dstip(sItem.dstip());
    }

    /* do work */
    if (!(tracebuf = new char[BIGBUF*batchSize])) {
	severe("out of memory in %s\n", __func__);
	goto done;
    }
    if (!initTraceroute())
	goto done;
    testsRemain = sched.item_size();
    while (testsRemain > 0) {
        if (testsRemain < batchSize) 
            batchSize = testsRemain;
        testsRemain -= batchSize;
	printf("%d done, %d in progress, %d remaining\n", testsDone, batchSize, testsRemain);
        runTraceroute(&stages[stageid], &dests[testsDone], tracebuf, batchSize);
        printf("\n");
        for (int i = 0; i < batchSize; i++) {
            /* copy the correct trace string into place */
            report->mutable_item(testsDone+i)->set_text(tracebuf + BIGBUF*i);
        }
        testsDone += batchSize;
    }
    report->set_status(DONE);
    
done:
    /* clean up */
    if (tracebuf) delete[] tracebuf;
    if (dests) delete[] dests;
}

static bool validateAddr(const std::string &addr, IPv ipv, const char *where, int index = -1)
{
    if (addr.size() != addrlen(ipvtofamily(ipv))) {
	if (index < 0)
	    severe("Protocol error: invalid IPv%d address size %zd in %s\n",
		ipv, addr.size(), where);
	else
	    severe("Protocol error: invalid IPv%d address size %zd in %s[%d]\n",
		ipv, addr.size(), where, index);
	return false;
    }
    return true;
}

bool Session::run()
{
    PBStreamer::Result pbsresult;

    // generate ClientHello message
    req.Clear();
    req.set_ready(true);

    req.mutable_hello()->set_version(VERSION);
    req.mutable_hello()->set_os(OS);
    req.mutable_hello()->set_share_public(sharePublic == 1);
    req.mutable_hello()->set_share_remedy(shareRemedy == 1);
    req.mutable_hello()->set_sessionkey(sessionkey);

    req.mutable_hello()->add_type(SPOOF);
    if (ipv == IPv4) {
	req.mutable_hello()->add_type(TRACEFILTER);
	req.mutable_hello()->add_type(TRACEROUTE);
#if 0
	req.mutable_hello()->add_type(SPOOFINGRESS);
#endif
    }

    if (!serverConnect()) return false;
    req.Clear();

    while (true) {
	// get server response
	if (standaloneMode) {
	    resp.Clear();
	    if (!rcvdHello) {
		generateFakeHello();
	    } else {
		// generate fake ResultSummary
		resp.mutable_summary()->set_clientasn(ipv==IPv4 ? 444 : 66666);
		resp.mutable_summary()->set_privaddr(ipv==IPv4 ?
		    SpooferResultSummary::BLOCKED : SpooferResultSummary::REWRITTEN);
		resp.mutable_summary()->set_routable(ipv==IPv4 ?
		    SpooferResultSummary::RECEIVED : SpooferResultSummary::UNKNOWN);
	    }
	} else {
	    if ((pbsresult = pbs->readFullMessage(&resp)) != PBStreamer::OK) {
		severe("receiving ServerMsg: %s\n", pbs->last_errmsg());
		return false;
	    }
	}
	printServerMsg(resp, ipv);
	if (!validateServerMsg())
	    return false;
	if (resp.has_summary()) // end of session
	    break;

	// prepare the scheduled tests
	cost = 0;
	progress = 0;
	for (stageid = 0; stageid < resp.schedule_size(); stageid++) {
	    const SpooferServerSchedule &sched = resp.schedule(stageid);
	    // note: we avoid the oneof API for 2.4 compatibility
	    if (sched.has_spoof()) {
		stages.push_back(Stage(this, sched.spoof()));
	    } else if (sched.has_tracefilter()) {
		stages.push_back(Stage(this, sched.tracefilter()));
	    } else if (sched.has_traceroute()) {
		stages.push_back(Stage(this, sched.traceroute()));
#if 0
	    } else if (sched.has_spoofingress()) {
		stages.push_back(Stage(this, sched.spoofingress()));
#endif
	    } else {
		severe("impossible: Unknown or missing schedule\n");
	    }
	}

	// NB: GUI depends on this output format
	printf(">> Estimated IPv%d probe count: %d\n", ipv, cost);

	// run the scheduled tests
	for (stageid = 0; stageid < resp.schedule_size(); stageid++) {
	    const SpooferServerSchedule &sched = resp.schedule(stageid);
	    // note: we avoid the oneof API for 2.4 compatibility
	    if (sched.has_spoof()) {
		runSpoofTests(sched.spoof());
	    } else if (sched.has_tracefilter()) {
		runTracefilterTests(sched.tracefilter());
	    } else if (sched.has_traceroute()) {
		runTracerouteTests(sched.traceroute());
#if 0
	    } else if (sched.has_spoofingress()) {
		runSpoofingressTests(sched.spoofingress());
#endif
	    }
	    schedulesDone++;
	}

	// send reports and request more work
	req.set_ready(true);
	printClientMsg(req, ipv);
	if (!standaloneMode) {
	    if ((pbsresult = pbs->writeMessage(&req)) != PBStreamer::OK) {
		severe("sending ClientMsg: %s\n", pbs->last_errmsg());
		return false;
	    }
	}
	req.Clear();
	stages.clear();
    }

    printf(">> IPv%d Result Summary:\n", ipv);
    printf(">>   ASN: %d\n", resp.summary().clientasn());
    printf(">>   Spoofed private addresses: %s\n",
	summaryResultToStr(resp.summary().privaddr()));
    printf(">>   Spoofed routable addresses: %s\n",
	summaryResultToStr(resp.summary().routable()));

    return true;
}

void Session::generateFakeHello()
{
    resp.Clear();

    std::vector<const char *> spoof_dests;
    char dst_in[IPV6ADDRLEN], src_in[IPV6ADDRLEN];

#if 0
    if (ipv == IPv4) {
	resp.mutable_txtmsg()->set_level(SpooferTextMessage::WARNING);
	resp.mutable_txtmsg()->set_body("This is a test of a SpooferTextMessage warning.  This is only a test.");
    } else {
	resp.mutable_txtmsg()->set_level(SpooferTextMessage::NOTICE);
	resp.mutable_txtmsg()->set_body("This is a test of a SpooferTextMessage notice.  This is only a test.");
    }
#endif

    if (ipv == IPv4) {
	spoof_dests.push_back("139.91.90.6");
	spoof_dests.push_back("78.41.116.2");
	// spoof_dests.push_back("137.164.84.50");
    } else {
	spoof_dests.push_back("2a02:60:1:1::49"); // vie-at.ark
	spoof_dests.push_back("2001:770:18:7:225:90ff:fe0c:acb4"); // dub-ie.ark
    }
    resp.mutable_hello()->set_sessionid(0xFFFFFFFF); // higher than any real id
    resp.mutable_hello()->set_sessionkey("FakeSessionKey"); // no caps in real keys
    resp.mutable_hello()->set_clientip(int_clientip);

    // initialize fake Spoof tests
    // (with separate Schedules for non-spoofed and spoofed phases)
    inet_pton(family, ipv==IPv4 ? "10.68.68.68" : "2001::1", &src_in);
    for (int spoof = 0; spoof <= 1; spoof++) {
	char seqno[SEQNOSIZE];
	SpooferSpoofSchedule *sched = resp.add_schedule()->mutable_spoof();
	SpooferSpoofSchedule::Item *item;
	for (unsigned i = 0; i < spoof_dests.size(); i++) {
	    item = sched->add_item();
	    if (spoof)
		item->set_srcip(&src_in, addrlen());
	    else
		item->set_srcip(resp.hello().clientip());
	    inet_pton(family, spoof_dests[i], &dst_in);
	    item->set_dstip(&dst_in, addrlen());
	    genSequence(seqno, SEQNOSIZE);
	    memcpy(seqno, "TEST", 4); // real seqno's never contain capitals
	    item->set_seqno(seqno, SEQNOSIZE);
	}
    }

    // initialize fake v4 Tracefilter tests
    if (ipv == IPv4) {
	SpooferTracefilterSchedule *sched = resp.add_schedule()->mutable_tracefilter();
	SpooferTracefilterSchedule::Item *item = sched->add_item();
	in_addr addr;
	inet_pton(AF_INET, standaloneTraceSrc ? standaloneTraceSrc : spoof_dests[0], &addr);
	item->set_srcip((char*)&addr, addrlen());
	((uint8_t*)&addr)[addrlen()-1] ^= 1; // toggle last bit
	item->set_dstip((char*)&addr, addrlen());
	item->set_dist(5);
    }

    // initialize fake v4 Traceroute tests
    if (ipv == IPv4) {
	SpooferTracerouteSchedule *sched = resp.add_schedule()->mutable_traceroute();
	SpooferTracerouteSchedule::Item *item;
	for (unsigned i = 0; i < spoof_dests.size(); i++) {
	    item = sched->add_item();
	    inet_pton(family, spoof_dests[i], &dst_in);
	    item->set_dstip(&dst_in, addrlen());
	}
    }
}

bool Session::validateServerMsg()
{
    if (resp.has_txtmsg() && !handleTextMessage(resp.txtmsg()))
	return false;

    if (resp.has_hello()) {
	if (rcvdHello) {
	    severe("unexpected ServerHello\n");
	    return false;
	}
	rcvdHello = true;
	if (!validateAddr(resp.hello().clientip(), ipv, "ServerHello"))
	    return false;
	ext_clientip = resp.hello().clientip();
	sessionid = resp.hello().sessionid();
	sessionkey = resp.hello().sessionkey();

    } else if (!rcvdHello) {
	severe("missing ServerHello\n");
	return false;
    }

    for (stageid = 0; stageid < resp.schedule_size(); stageid++) {
	int n_sched = 0;

	// note: we avoid the oneof API for 2.4 compatibility
	if (resp.schedule(stageid).has_spoof()) {
	    n_sched++;
	    const SpooferSpoofSchedule &sched = resp.schedule(stageid).spoof();
	    if (sched.item_size() > 400) {
		severe("Bad spoof test list from server (size=%d).\n",
		    sched.item_size());
		return false;
	    }
	    for (int j = 0; j < sched.item_size(); j++) {
		const SpooferSpoofSchedule::Item &item = sched.item(j);
		if (!validateAddr(item.srcip(), ipv, "SpoofSchedule", j))
		    return false;
		if (!validateAddr(item.dstip(), ipv, "SpoofSchedule", j))
		    return false;
		// XXX validate seqno, timestamp, hmac?
	    }
	}

	if (resp.schedule(stageid).has_tracefilter()) {
	    n_sched++;
	    const SpooferTracefilterSchedule &sched = resp.schedule(stageid).tracefilter();
	    if (sched.item_size() > 100) {
		warn("Ignoring bad tracefilter test list from server (size=%d).\n",
		    sched.item_size());
		resp.mutable_schedule(stageid)->mutable_tracefilter()->Clear();
	    }
	    for (int j = 0; j < sched.item_size(); j++) {
		const SpooferTracefilterSchedule::Item &item = sched.item(j);
		if (!validateAddr(item.srcip(), ipv, "TracefilterSchedule", j))
		    return false;
		if (!validateAddr(item.dstip(), ipv, "TracefilterSchedule", j))
		    return false;
	    }
	}

	if (resp.schedule(stageid).has_traceroute()) {
	    n_sched++;
	    const SpooferTracerouteSchedule &sched = resp.schedule(stageid).traceroute();
	    if (sched.item_size() > 100) {
		warn("Ignoring bad traceroute test list from server (size=%d).\n",
		    sched.item_size());
		resp.mutable_schedule(stageid)->mutable_traceroute()->Clear();
	    }
	    for (int j = 0; j < sched.item_size(); j++) {
		const SpooferTracerouteSchedule::Item &item = sched.item(j);
		if (!validateAddr(item.dstip(), ipv, "TracerouteSchedule", j))
		    return false;
	    }
	}

	if (n_sched == 0) {
	    severe("Schedule %d: Unknown or missing sub-schedule\n", stageid);
	    return false;
	} else if (n_sched > 1) {
	    severe("Schedule %d: multiple sub-schedules (%d)\n", stageid, n_sched);
	    return false;
	}
    }

    if (resp.has_summary() && resp.schedule_size() > 0)
	warn("Found %d schedules in final message\n", resp.schedule_size());

    return true;
}

bool Session::init()
{
    if (pretendMode) return true;

    bool result = initSockets(true, NULL);
    if (!result) {
	if (spoof_layer[ipv])
	    notice("Can not test IPv%d spoofing at layer %d.\n", ipv, spoof_layer[ipv]);
	else
	    notice("Can not test IPv%d spoofing.\n", ipv);
    }
    resetSockets();
    return result;
}

bool Session::serverConnect()
{
    struct probe_info *server_pinfo = 0;
    socklen_t len = 0;

    if (standaloneMode) {
	char ip[IPV6ADDRLEN];
	inet_pton(family, ipv == IPv4 ? "127.0.0.1" : "::1", ip);
	int_clientip.assign(ip, addrlen());
	req.mutable_hello()->set_clientip(int_clientip);
	printClientMsg(req, ipv);
	return true;
    }

    /* Establish TCP control connection */
    printf(">> Connecting to IPv%d server...\n", ipv);
    sockaddr_storage server_ss;
    struct sockaddr *server_sa = (struct sockaddr *)&server_ss;
    sockaddr_storage client_ss;
    struct sockaddr *client_sa = (struct sockaddr *)&client_ss;

    memset(server_sa, 0, sizeof(server_ss));
    server_sa->sa_family = family;
    inet_pton(family, ipv == IPv4 ? server4 : server6, sa_ipaddr(server_sa));
    sa_port(server_sa) = htons(REPORT_PORT);
    ctrlsock = newSocket(family, SOCK_STREAM, IPPROTO_TCP);
    if (ctrlsock < 0) {
	severe("can't create server socket.\n");
	goto fail;
    }
    debug(DEVELOP, "created ctrlsock\n");
    if (connect(ctrlsock, server_sa, sa_len(server_sa)) < 0) {
        severe("connecting to %s port %d: %s\n", sa_ntop(server_sa, NULL, 0),
	    ntohs(sa_port(server_sa)), sock_last_errmsg());
	goto fail;
    }
    debug(DEVELOP, "connected to %s port %d\n", sa_ntop(server_sa, NULL, 0),
	ntohs(sa_port(server_sa)));

    server_pinfo = new_probe_info(IPPROTO_TCP, false, ctrlsock, NULL, server_sa);

    len = sizeof(client_ss);
    if (getsockname(ctrlsock, client_sa, &len) < 0) {
        severe("Couldn't get IP address: %s", sock_last_errmsg());
	goto fail;
    } 
    int_clientip.assign(sa_ipaddr(client_sa), addrlen());
    req.mutable_hello()->set_clientip(int_clientip);

    printClientMsg(req, ipv);
    pbs = new PBStreamer(ctrlsock, PBS_MAGIC, sizeof(PBS_MAGIC));

    /* Say Hello. */
    PBStreamer::Result pbsresult;
    if ((pbsresult = pbs->writeMessage(&req)) != PBStreamer::OK) {
	severe("sending SpooferClientHello: %s\n", pbs->last_errmsg());
	goto fail;
    }
    return true;

fail:
    if (server_pinfo) free_probe_info(server_pinfo);
    notice("Can not test IPv%d spoofing.\n", ipv);
    if (ctrlsock >= 0) close(ctrlsock);
    ctrlsock = -1;
    return false;
}

static void displayResults(const char *skey) {
    char buf[BUFSIZE];
    if (!*skey) return;
    snprintf(buf, BUFSIZE, "http://%s/report.php?sessionkey=%s", 
        report_host, skey);
    printf("\nTest Complete.\n");
    // NB: GUI depends on this output format
    printf("Your test results:\n    %s\n\n", buf);
#if 0 // don't open webbrowser
    if (is_interactive()) {
	// open webbrowser to display report
#ifdef _WIN32
	debug(DEVELOP, ">> prober: launching browser\n");
	ShellExecute(NULL, "open", buf, NULL, NULL, SW_SHOWNORMAL);
#else
 #ifdef _OSX
	const char *opencmd = "open";
 #else // unix gui?
	const char *opencmd = "xdg-open";
 #endif
	debug(DEVELOP, ">> prober: launching browser\n");
	execlp(opencmd, opencmd, buf, NULL);
#endif
    }
#endif
}

static void banner() {
    // NB: GUI depends on this output format
    printf(">> %s v%s\n", PACKAGE_DESC, PACKAGE_VERSION);
    printf(">> %s\n", PACKAGE_URL);
    const char *start;
    int len;
    for (start = COPYRIGHT, len = 0; start[len]; start += len + 1) {
	while (isspace(*start)) ++start;
	len = strcspn(start, ";");
	if (len > 0) printf(">> %.*s\n", len, start);
    }
    printf(">>\n");

    printf(">> %s\n", pcap_lib_version());
    printf(">> Google Protobuf version %d.%d.%d\n",
	GOOGLE_PROTOBUF_VERSION / 1000000,
	GOOGLE_PROTOBUF_VERSION / 1000 % 1000,
	GOOGLE_PROTOBUF_VERSION % 1000);
    printf(">>\n");

    for (start = PACKAGE_LONGDESC, len = 0; start[0]; start += len) {
	while (isspace(start[len])) ++start;
	if ((len = strlen(start)) > 75) {
	    len = 75;
	    while (!isspace(start[len-1])) --len;
	}
	printf(">> %.*s\n", len, start);
    }
    printf(">>\n");
}

static NORETURN usage(char *prog, int exitcode) {
    banner();
    printf("\n");
    printf("Usage %s [options]\n", prog);
    printf("options:\n");
    printf("-h, --help  print this help and exit\n");
    printf("--version   print version information and exit\n");
    printf("-s{1|0}     " DESC_SHARE_PUBLIC " (yes/no)\n");
    printf("-r{1|0}     " DESC_SHARE_REMEDY " (yes/no)\n");
    printf("-v          be more verbose (repeatable)\n");
    printf("-4          test IPv4 (default: both IPv4 and IPv6)\n");
    printf("-6          test IPv6 (default: both IPv4 and IPv6)\n");
#if DEBUG
    printf("-P          pretend mode - don't send any probe packets\n");
    printf("-S          standalone debugging mode - run a test without server\n");
    printf("-T          use the development test server\n");
    printf("-l<n>       spoof at layer <n>: 3 (IP), 2 (link), or 0 (auto).\n");
#endif
    printf("\n");
    exitpause(exitcode);
}

static void optionPrompt(int &opt, const char *name, const char *desc)
{
    if (opt < 0) {
	char buf[SMALLBUF];
	if (!is_interactive())
	    fatal("Missing required %s option.\n", name);
	printf("%s? [yes/no/quit] ", desc);
	if (standaloneMode) printf("\n(ignored in standalone mode) ");
	fflush(stdout);
	if (fgets(buf, SMALLBUF, stdin) == NULL)
	    fatal("Aborting.\n");
	if (tolower(buf[0]) == 'y')
	    opt = 1;
	else if (tolower(buf[0]) == 'n')
	    opt = 0;
	else
	    fatal("Aborting.\n");
    }
}

int main(int argc, char **argv) {
    int opt;
    bool enable4 = false, enable6 = false;

    setvbuf(stderr, NULL, _IONBF, 0);
    setvbuf(stdout, NULL, _IONBF, 0);

    while ((opt = getopt(argc, argv, "s:r:vPSThl:46-:")) != -1) {
        switch (opt) {
        case 's':
	    if (strcmp(optarg, "1") == 0)
		sharePublic = 1;
	    else if (strcmp(optarg, "0") == 0)
		sharePublic = 0;
	    else
		usage(argv[0], -1);
            break;
        case 'r':
	    if (strcmp(optarg, "1") == 0)
		shareRemedy = 1;
	    else if (strcmp(optarg, "0") == 0)
		shareRemedy = 0;
	    else
		usage(argv[0], -1);
            break;
        case 'v':
            verbosity++;
            break;
        case 'P':
            pretendMode = TRUE;
            break;
        case 'S':
            standaloneMode = TRUE;
            break;
        case 'T':
	    useDevServer = true;
            break;
        case 'l':
            spoof_layer[IPv4] = atoi(optarg);
            spoof_layer[IPv6] = atoi(optarg);
            break;
	case '4':
	    enable4 = true;
	    break;
	case '6':
	    enable6 = true;
	    break;
        case 'h':
            usage(argv[0], 0); // no return
        case '-': // hack (accepts "- -version" as well as "--version")
	    if (strcmp(optarg, "version") == 0) {
		banner();
		exit(0);
	    } else if (strcmp(optarg, "help") == 0) {
		usage(argv[0], 0); // no return
	    }
	    fprintf(stderr, "%s: invalid long option: '%s'\n", argv[0], optarg);
	    // fall through
        default:
            usage(argv[0], -1); // no return
        }
    }

    banner();

    if (useDevServer) {
	server4 = TEST_SERVER;
	server6 = TEST_SERVER6;
	report_host = TEST_REPORT_HOST;
    }

    if (!enable4 && !enable6)
	enable4 = enable6 = true; // default to both

    /* Init */
    streamSetup();
    srand(time(NULL));

    /* Ask about data sharing */
    printf(">> Ready to test (please allow several minutes to complete).\n");

    optionPrompt(sharePublic, "-s", DESC_SHARE_PUBLIC);
    optionPrompt(shareRemedy, "-r", DESC_SHARE_REMEDY);

    printf(">> Options:\n");
    printf(">>   IPv4: %s\n", enable4 ? "enabled" : "disabled");
    printf(">>   IPv6: %s\n", enable6 ? "enabled" : "disabled");
    printf(">>   sharePublic: %s\n", sharePublic ? "yes" : "no");
    printf(">>   shareRemedy: %s\n", shareRemedy ? "yes" : "no");
    if (useDevServer) printf(">>   useDevServer\n");
    if (pretendMode) printf(">>   pretendMode\n");
    if (standaloneMode) printf(">>   standaloneMode\n");

#ifdef _WIN32
    win_init();
#endif

#ifdef REGRESS 
  #include "regress.c"
#endif 

    SessionStatus status4 = SS_UNTESTED, status6 = SS_UNTESTED;

    // Notes:
    // * If both v4 and v6 are enabled, the server expects v4 first.
    // * Failure in init() is not something that is likely to change in the
    //   near future, so we do not indicate it in exitstatus, so the Scheduler
    //   will not retry this Prober.
    if (enable4) {
	Session session4(IPv4);
	if (session4.init()) {
	    status4 = session4.run() ? SS_SUCCESS : SS_FAIL;
	}
    }
    if (enable6) {
	Session session6(IPv6);
	if (session6.init()) {
	    status6 = session6.run() ? SS_SUCCESS : SS_FAIL;
	}
    }

    displayResults(sessionkey.c_str());
    exitpause((std::max(status4, status6) >= SS_SUCCESS) ? 0 : 1);
}
