/* 
 * Copyright 2004-2009 Rob Beverly
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

/****************************************************************************
   Author:      Rob Beverly <rbeverly@mit.edu>
   Date:        $Date: 2016/05/26 17:26:34 $
   Description: WIN32-Specific traceroute wrapper
****************************************************************************/
#ifdef _WIN32

#include "spoof.h"
#include "prober.h"
static const char cvsid[] USED = "$Id: wintrace.cc,v 1.14 2016/05/26 17:26:34 kkeys Exp $";

#define TRWAIT 60  // wait at most this long for traceroute to finish
#define TRDEBUG 0

HANDLE ghMutex;
int children_living = 0;
HANDLE gStdOut_Rd[MAX_TRACEROUTE_CHILD];
HANDLE gStdOut_Wr[MAX_TRACEROUTE_CHILD];
int progress[MAX_TRACEROUTE_CHILD];
int shownProgress[MAX_TRACEROUTE_CHILD];

/* thread-specific data */
typedef struct MyData {
    int id;
    struct in_addr *dest;
    char *result;
} MYDATA, *PMYDATA; 


/* thread which performs tracing work */
static DWORD WINAPI traceThread(LPVOID lpParam) {
    PMYDATA pDataArray;
    char tracecmd[BUFSIZE];
    int bytes_read = 0;

    /* from: http://msdn.microsoft.com/en-us/library/ms682499.aspx */
    STARTUPINFO siStartInfo;
    PROCESS_INFORMATION piProcInfo;
    SECURITY_ATTRIBUTES saAttr;
    DWORD dwRead;
    BOOL bSuccess = FALSE;

    pDataArray = (PMYDATA) lpParam;
#if TRDEBUG
    printf(">> Running Trace to: %s [thread: %d]\n", 
        inet_ntoa(*(pDataArray->dest)), pDataArray->id);
#endif
    sprintf(tracecmd, "tracert -d %s", inet_ntoa(*(pDataArray->dest)));

    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

    CreatePipe(&gStdOut_Rd[pDataArray->id], &gStdOut_Wr[pDataArray->id], &saAttr, 0);
    SetHandleInformation(gStdOut_Rd[pDataArray->id], HANDLE_FLAG_INHERIT, 0);

    ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
    siStartInfo.cb = sizeof(STARTUPINFO);
    siStartInfo.hStdError = NULL;
    siStartInfo.hStdOutput = gStdOut_Wr[pDataArray->id];
    siStartInfo.dwFlags |= STARTF_USESTDHANDLES;

    CreateProcess(NULL, tracecmd, NULL, NULL, TRUE, 0,
        NULL, NULL, &siStartInfo, &piProcInfo);
    CloseHandle(gStdOut_Wr[pDataArray->id]);
    while (1) {
        if (children_living == 0) {
#if TRDEBUG
            printf("I should die [thread: %d]\n", pDataArray->id);
#endif
            ExitThread(0);
        }

	if (bytes_read >= BIGBUF - 1) // buffer full
	    break;
	bSuccess = ReadFile(gStdOut_Rd[pDataArray->id], 
	    pDataArray->result + bytes_read, BIGBUF - bytes_read - 1, &dwRead, NULL);
	if (!bSuccess) // error (including EOF)
	    break;
	if (dwRead > 0) {
	    for (unsigned j = 0; j < dwRead; j++) {
		// Show a progress tick for each non-empty line of
		// traceroute output.
		if ((bytes_read > 0 || j > 0) &&
		    pDataArray->result[bytes_read + j] == '\n' &&
		    pDataArray->result[bytes_read + j - 1] != '\n' &&
		    progress[pDataArray->id] < TRACEROUTE_ITEM_COST - 1)
		{
		    progress[pDataArray->id]++;
		}
	    }
	    bytes_read += dwRead;
        }
#if TRDEBUG
        printf("Thread: %d (%p) Read: [%lu] Total: [%d]\n",
                pDataArray->id, pDataArray, dwRead, bytes_read);
#endif
	if (dwRead <= 0)
            Sleep(200);
    }
    pDataArray->result[bytes_read] = '\0';
#if TRDEBUG
    printf("Thread %d exit.\n", pDataArray->id);
#endif
    WaitForSingleObject(ghMutex, INFINITE);
    progress[pDataArray->id] = TRACEROUTE_ITEM_COST; 
    children_living--;
    ReleaseMutex(ghMutex);
    ExitThread(0);
    return 0;
}

int initTraceroute() { return 1; }

/**
 * runTraceroute - perform a traceroute to a destination 
 *
 * @param[in]   dest        destination
 * @param[out]  result      traceroute output
 * @param[in]   num     number of destinations
 *
 * Performs a traceroute to a destination.  Uses tracert
 * binary, starts time-limited trace.  Will terminate
 * after TRWAIT seconds
 */
int runTraceroute(Tracker *tracker, struct in_addr *dest, char *result, int num) {
    PMYDATA pDataArray[MAX_TRACEROUTE_CHILD];
    HANDLE hThreadArray[MAX_TRACEROUTE_CHILD];
    int lastStatus, started, now;
    int i;

    if (num > MAX_TRACEROUTE_CHILD)
        num = MAX_TRACEROUTE_CHILD;

    /* Initialize thread data */
    memset(result, 0, num*BIGBUF); // guarantees NUL-termination even if thread is killed
    for (i=0;i<num;i++) {
        pDataArray[i] = (PMYDATA) HeapAlloc(GetProcessHeap(),
            HEAP_ZERO_MEMORY, sizeof(MYDATA));
        pDataArray[i]->id = i;
        pDataArray[i]->dest = dest + i;
        pDataArray[i]->result = result + (i*BIGBUF); 
        children_living++;
        progress[i] = -1; // start at -1 to skip the first line of output
        shownProgress[i] = 0;
    }
    ghMutex = CreateMutex(NULL, FALSE, NULL);

    /* Start threads */
    started = lastStatus = time(NULL);
    for (i=0;i<num;i++) {
        hThreadArray[i] = CreateThread(NULL, 0, traceThread, 
            pDataArray[i], 0, NULL);
    }
    
    /* Wait for threads to finish, or timeout */
    while (children_living > 0) {
        Sleep(200);
	now = time(NULL);
	for (i=0;i<num;i++) {
	    while (shownProgress[i] < progress[i]) {
		tracker->showProgress();
		shownProgress[i]++;
	    }
	}
#if TRDEBUG
        printf("%d threads active\n", children_living);
#endif
	if (now - started > TRWAIT) {
            printf("finishing.");
            WaitForSingleObject(ghMutex, INFINITE);
            children_living = 0;
            ReleaseMutex(ghMutex);
        }
    }

    /* Shut down threads */
    Sleep(1000);
    for (i=0;i<num;i++) {
	if (WaitForSingleObject(hThreadArray[i], 0) != WAIT_OBJECT_0) {
#if TRDEBUG
            printf("Thread %d still running. shut it down.\n", i);
#endif
	    TerminateThread(hThreadArray[i], 666);
	    WaitForSingleObject(hThreadArray[i], INFINITE);
#if TRDEBUG
	    printf("Thread %d terminated.\n", i);
#endif
	}
        CloseHandle(hThreadArray[i]);
        if (pDataArray[i] != NULL) {
            HeapFree(GetProcessHeap(), 0, pDataArray[i]);
            pDataArray[i] = NULL;
        }
	while (shownProgress[i] < TRACEROUTE_ITEM_COST) {
	    tracker->showProgress();
	    shownProgress[i]++;
	}
    }
    return 1;
}

#if 0
int main () {
    struct in_addr *dests;
    char *result;
    int numdests = 5;
    int i;

    result = calloc(numdests, BIGBUF);
    dests  = calloc(numdests, sizeof(struct in_addr));

    dests[0].s_addr = inet_addr("192.172.226.5");
    dests[1].s_addr = inet_addr("128.61.2.1");
    dests[2].s_addr = inet_addr("128.63.2.52");
    dests[3].s_addr = inet_addr("192.168.1.1");
    dests[4].s_addr = inet_addr("128.89.80.99");
//    dests[0].s_addr = inet_addr("64.233.161.99");
//    dests[1].s_addr = inet_addr("18.7.22.83");
//    dests[2].s_addr = inet_addr("128.61.2.1");

    runtrace(dests, result, numdests);
    for (i=0;i<numdests;i++) {
        printf("WinTrace %d: %s\n", i, &result[BIGBUF*i]);
    }
    exit(0);
}
#endif

#endif
